/* chi-fast64.c: CHI hash 64 bit optimized implementation. */

/*
THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE AND AGAINST
INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <string.h>
#include <stdio.h>

#include "chi.h"

#if defined(_WIN32) || defined(_WIN64)
#   define bswap_64 _byteswap_uint64
#else
#   include <byteswap.h>
#endif

/*
 * Regardless of endian-ness converts a byte to / from a big endian array of
 * bytes.
 */
#ifdef _BIG_ENDIAN
#    define BYTE2WORD(b) (*(WORD64 *)(b))
#else
#    define BYTE2WORD(b) bswap_64(*(WORD64 *)(b))
#endif

/*
 * Rotate WORD32 to the right by r
 */
static INLINE WORD32
rotr32
(
    WORD32  X,
    int     r
)
{
    return (X >> r) | (X << (32 - r));
}

/*
 * Rotate WORD64 to the right by r
 */
static INLINE WORD64
rotr64
(
    WORD64  X,
    int     r
)
{
    return (X >> r) | (X << (64 - r));
}

/*
 * Rotate Upper and Lower 32 bits of WORD64 to
 * the right by r1 and r2 respectively.
 */
static INLINE WORD64
drotr32
(
    WORD64  X,
    int     r1,
    int     r2
)
{
    return (WORD64)rotr32((WORD32)(X >> 32), r1) << 32 | rotr32((WORD32)X, r2);
}

/*
 * Byte swap
 */
static INLINE WORD64
swap8
(
    WORD64 X
)
{
    return bswap_64(X);
}

/*
 * Swap the Upper and Lower 32 bits of WORD64
 */
static INLINE WORD64
swap32
(
    WORD64 X
)
{
    return rotr64(X, 32);
}

/*
 * Non-linear MAP function.
 */
static INLINE void
map
(
    WORD64  R,
    WORD64  S,
    WORD64  T,
    WORD64  U,
    WORD64  *X
)
{
    WORD64  i;

    i    = (T | U) ^ R;
    X[0] = (i &  S) | ((~T ^ (U & R)) & ~S);
    X[1] = (i & ~S) | (( U ^ (R & T)) &  S);
    X[2] = ( U & (S ^ (T & ~R))) | (~U & (R ^ (T &  S)));
}

/*
 * CHI-256
 */
#define _256_MAP map

static INLINE WORD64
_256_theta0
(
    WORD64  X
)
{
    WORD32  U, L;

    U = (WORD32)(X >> 32);
    L = (WORD32)X;
    U = rotr32(U, 21) ^ rotr32(U, 26) ^ rotr32(U, 30);
    L = rotr32(L, 21) ^ rotr32(L, 26) ^ rotr32(L, 30);

    return (WORD64)U << 32 | L;
}

static INLINE WORD64
_256_theta1
(
    WORD64  X
)
{
    WORD32  U, L;

    U = (WORD32)(X >> 32);
    L = (WORD32)X;
    U = rotr32(U, 14) ^ rotr32(U, 24) ^ rotr32(U, 31);
    L = rotr32(L, 14) ^ rotr32(L, 24) ^ rotr32(L, 31);

    return (WORD64)U << 32 | L;
}

static INLINE WORD64
_256_mu0
(
    WORD64  X
)
{
    return rotr64(X, 36) ^ rotr64(X, 18) ^ (X >> 1);
}

static INLINE WORD64
_256_mu1
(
    WORD64  X
)
{
    return rotr64(X, 59) ^ rotr64(X, 37) ^ (X >> 10);
}

static INLINE void
_256_premix
(
    WORD64  A,
    WORD64  B,
    WORD64  D,
    WORD64  E,
    WORD64  *R,
    WORD64  *S,
    WORD64  *T,
    WORD64  *U
)
{
    WORD64 dswap;

    *T = drotr32(swap32(A), 7, 26) ^ drotr32(D, 14, 22);
    dswap = swap32(D);
    *R = drotr32(B, 8, 8)          ^ drotr32(dswap, 5, 1);
    *S = A                         ^ drotr32(dswap, 18, 17);
    *U = drotr32(A, 17, 12)        ^ drotr32(swap32(E), 2, 23);
}

static INLINE void
_256_datainput
(
    WORD64  V0,
    WORD64  V1,
    WORD64  *R,
    WORD64  *S,
    WORD64  *T,
    WORD64  *U
)
{
    *R ^= V0;
    *S ^= V1;
    *T ^= _256_theta0(V0);
    *U ^= _256_theta1(V1);
}


static INLINE void
_256_premix_and_di
(
    WORD64  A,
    WORD64  B,
    WORD64  D,
    WORD64  E,
    WORD64  V0,
    WORD64  V1,
    WORD64  *R,
    WORD64  *S,
    WORD64  *T,
    WORD64  *U
)
{
    WORD64 dswap;

    *T = drotr32(D, 14, 22) ^ drotr32(swap32(A), 7, 26) ^ _256_theta0(V0);
    dswap = swap32(D);
    *R = drotr32(B,  8,  8) ^ drotr32(dswap,     5,  1) ^ V0;
    *S = A                  ^ drotr32(dswap,    18, 17) ^ V1;
    *U = drotr32(A, 17, 12) ^ drotr32(swap32(E), 2, 23) ^ _256_theta1(V1);
}

static INLINE void
_256_postmix
(
    WORD64  X,
    WORD64  Y,
    WORD64  Z,
    WORD64  *F,
    WORD64  *C
)
{
    WORD64  YY;

    YY = swap8(drotr32(Y, 5, 5));
    *F ^= rotr64(X + YY, 16);
    *C ^= rotr64(swap32(Z) + YY, 48);
}

#define _256_STEP(A, B, C, D, E, F, i)                                         \
    do {                                                                       \
        _256_premix(A, B, D, E, &R, &S, &T, &U);                               \
        _256_datainput(W[2*(i)]^K[2*(i)], W[2*(i)+1]^K[2*(i)+1], &R,&S,&T,&U); \
        _256_MAP(R, S, T, U, M);                                               \
        _256_postmix(M[(i)%3], M[((i)+1)%3], M[((i)+2)%3], &F, &C);            \
    } while(0)

#define _256_STEP_2(A, B, C, D, E, F, i)                                       \
    do {                                                                       \
        _256_STEP(A, B, C, D, E, F, (i) + 0);                                  \
        _256_STEP(F, A, B, C, D, E, (i) + 1);                                  \
    } while(0)

#define _256_STEP_6(A, B, C, D, E, F, i)                                       \
    do {                                                                       \
        _256_STEP(A, B, C, D, E, F, (i) + 0);                                  \
        _256_STEP(F, A, B, C, D, E, (i) + 1);                                  \
        _256_STEP(E, F, A, B, C, D, (i) + 2);                                  \
        _256_STEP(D, E, F, A, B, C, (i) + 3);                                  \
        _256_STEP(C, D, E, F, A, B, (i) + 4);                                  \
        _256_STEP(B, C, D, E, F, A, (i) + 5);                                  \
    } while(0)

#define _256_MSG(W, i) (W[(i)-8] ^ _256_mu0(W[(i)-7]) ^ _256_mu1(W[(i)-2]))

#define _256_MSG_EXP(W, b)                                                     \
    do {                                                                       \
        W[0] = BYTE2WORD(b + 8*0);                                             \
        W[1] = BYTE2WORD(b + 8*1);                                             \
        W[2] = BYTE2WORD(b + 8*2);                                             \
        W[3] = BYTE2WORD(b + 8*3);                                             \
        W[4] = BYTE2WORD(b + 8*4);                                             \
        W[5] = BYTE2WORD(b + 8*5);                                             \
        W[6] = BYTE2WORD(b + 8*6);                                             \
        W[7] = BYTE2WORD(b + 8*7);                                             \
        for (i = 8; i < 2 * _256_STEPS; i++)                                   \
            W[i] = _256_MSG(W, i);                                             \
    } while(0)

/*
 * Parameters:
 *
 *      state   The hash state.
 *      final   1 for the final block in the hash, 0 for intermediate blocks.
 */
void
_256_update
(
    hashState   *state,
    int         final
)
{
    int         i;
    WORD64      W[2*_256_STEPS];

    WORD64      A, B, C, D, E, F;
    WORD64      R, S, T, U;
    WORD64      M[3];

    if (final)
    {
        A = state->hs_State.small[0] = rotr64(state->hs_State.small[0], 1);
        B = state->hs_State.small[1] = rotr64(state->hs_State.small[1], 1);
        C = state->hs_State.small[2] = rotr64(state->hs_State.small[2], 1);
        D = state->hs_State.small[3] = rotr64(state->hs_State.small[3], 1);
        E = state->hs_State.small[4] = rotr64(state->hs_State.small[4], 1);
        F = state->hs_State.small[5] = rotr64(state->hs_State.small[5], 1);
    }
    else
    {
        state->hs_State.small[0] = rotr64(A = state->hs_State.small[0], 1);
        state->hs_State.small[1] = rotr64(B = state->hs_State.small[1], 1);
        state->hs_State.small[2] = rotr64(C = state->hs_State.small[2], 1);
        state->hs_State.small[3] = rotr64(D = state->hs_State.small[3], 1);
        state->hs_State.small[4] = rotr64(E = state->hs_State.small[4], 1);
        state->hs_State.small[5] = rotr64(F = state->hs_State.small[5], 1);
    }

    _256_MSG_EXP(W, state->hs_DataBuffer);

    _256_STEP_6(A, B, C, D, E, F,  0);
    _256_STEP_6(A, B, C, D, E, F,  6);
    _256_STEP_6(A, B, C, D, E, F, 12);
    _256_STEP_2(A, B, C, D, E, F, 18);

    state->hs_State.small[0] ^= E;
    state->hs_State.small[1] ^= F;
    state->hs_State.small[2] ^= A;
    state->hs_State.small[3] ^= B;
    state->hs_State.small[4] ^= C;
    state->hs_State.small[5] ^= D;
}

/*
 * CHI-512
 */
#define _512_MAP map

static INLINE WORD64
_512_theta0
(
    WORD64  X
)
{
    return rotr64(X,  5) ^ rotr64(X,  6) ^ rotr64(X, 43);
}

static INLINE WORD64
_512_theta1
(
    WORD64  X
)
{
    return rotr64(X, 20) ^ rotr64(X, 30) ^ rotr64(X, 49);
}

static INLINE WORD64
_512_mu0
(
    WORD64  X
)
{
    return rotr64(X, 36) ^ rotr64(X, 18) ^ (X >> 1);
}

static INLINE WORD64
_512_mu1
(
    WORD64  X
)
{
    return rotr64(X, 60) ^ rotr64(X, 30) ^ (X >> 3);
}

static INLINE void
_512_premix
(
    WORD64  A,
    WORD64  B,
    WORD64  D,
    WORD64  E,
    WORD64  G,
    WORD64  P,
    WORD64  *R,
    WORD64  *S,
    WORD64  *T,
    WORD64  *U
)
{
    *T = P;
    *S = A;
    *T ^= rotr64(A, 11);
    *U = rotr64(A, 26);
    *R = rotr64(B, 11);
    *R ^= rotr64(D, 8);
    *S ^= rotr64(D, 21);
    *T ^= rotr64(D, 38);
    *U ^= rotr64(E, 40);
    *R ^= rotr64(G, 13);
    *S ^= rotr64(G, 29);
    *U ^= rotr64(G, 50);
}

static INLINE void
_512_datainput
(
    WORD64  V0,
    WORD64  V1,
    WORD64  *R,
    WORD64  *S,
    WORD64  *T,
    WORD64  *U
)
{
    *R ^= V0;
    *S ^= V1;
    *T ^= _512_theta0(V0);
    *U ^= _512_theta1(V1);
}

static INLINE void
_512_postmix
(
    WORD64  X,
    WORD64  Y,
    WORD64  Z,
    WORD64  *Q,
    WORD64  *F,
    WORD64  *C
)
{
    WORD64  YY, ZZ;

    YY = swap8(rotr64(Y, 31));
    ZZ = swap32(Z);
    *Q ^= rotr64( X + YY, 16);
    *C ^= rotr64(ZZ + YY, 48);
    *F ^= ZZ + (X << 1);
}

#define _512_STEP(A, B, C, D, E, F, G, P, Q, i)                                \
    do {                                                                       \
        _512_premix(A, B, D, E, G, P, &R, &S, &T, &U);                         \
        _512_datainput(W[2*(i)]^K[2*(i)], W[2*(i)+1]^K[2*(i)+1],&R,&S,&T,&U);  \
        _512_MAP(R, S, T, U, M);                                               \
        _512_postmix(M[(i)%3], M[((i)+1)%3], M[((i)+2)%3], Q, F, C);           \
    } while(0)

#define _512_STEP_4(A, B, C, D, E, F, G, P, Q, i)                              \
    do {                                                                       \
        _512_STEP(A, B, &C, D, E, &F, G, P, &Q, (i) + 0);                      \
        _512_STEP(Q, A, &B, C, D, &E, F, G, &P, (i) + 1);                      \
        _512_STEP(P, Q, &A, B, C, &D, E, F, &G, (i) + 2);                      \
        _512_STEP(G, P, &Q, A, B, &C, D, E, &F, (i) + 3);                      \
    } while(0)

#define _512_STEP_9(A, B, C, D, E, F, G, P, Q, i)                              \
    do {                                                                       \
        _512_STEP(A, B, &C, D, E, &F, G, P, &Q, (i) + 0);                      \
        _512_STEP(Q, A, &B, C, D, &E, F, G, &P, (i) + 1);                      \
        _512_STEP(P, Q, &A, B, C, &D, E, F, &G, (i) + 2);                      \
        _512_STEP(G, P, &Q, A, B, &C, D, E, &F, (i) + 3);                      \
        _512_STEP(F, G, &P, Q, A, &B, C, D, &E, (i) + 4);                      \
        _512_STEP(E, F, &G, P, Q, &A, B, C, &D, (i) + 5);                      \
        _512_STEP(D, E, &F, G, P, &Q, A, B, &C, (i) + 6);                      \
        _512_STEP(C, D, &E, F, G, &P, Q, A, &B, (i) + 7);                      \
        _512_STEP(B, C, &D, E, F, &G, P, Q, &A, (i) + 8);                      \
    } while(0)

#define _512_MSG(W,i) (W[i-16] ^ W[i-7] ^ _512_mu0(W[i-15]) ^ _512_mu1(W[i-2]))

#define _512_MSG_EXP(W, b)                                                     \
    do {                                                                       \
        W[0] = BYTE2WORD(b + 8*0);                                             \
        W[1] = BYTE2WORD(b + 8*1);                                             \
        W[2] = BYTE2WORD(b + 8*2);                                             \
        W[3] = BYTE2WORD(b + 8*3);                                             \
        W[4] = BYTE2WORD(b + 8*4);                                             \
        W[5] = BYTE2WORD(b + 8*5);                                             \
        W[6] = BYTE2WORD(b + 8*6);                                             \
        W[7] = BYTE2WORD(b + 8*7);                                             \
        W[8] = BYTE2WORD(b + 8*8);                                             \
        W[9] = BYTE2WORD(b + 8*9);                                             \
        W[10] = BYTE2WORD(b + 8*10);                                           \
        W[11] = BYTE2WORD(b + 8*11);                                           \
        W[12] = BYTE2WORD(b + 8*12);                                           \
        W[13] = BYTE2WORD(b + 8*13);                                           \
        W[14] = BYTE2WORD(b + 8*14);                                           \
        W[15] = BYTE2WORD(b + 8*15);                                           \
        for (i = 16; i < 2 * _512_STEPS; i++)                                  \
            W[i] = _512_MSG(W, i);                                             \
    } while(0)

/*
 * Parameters:
 *      state   The hash state.
 *      final   1 for the final block in the hash, 0 for intermediate blocks.
 */
void
_512_update
(
    hashState   *state,
    int         final
)
{
    int         i;

    WORD64      W[2*_512_STEPS];
    WORD64      A, B, C, D, E, F, G, P, Q;
    WORD64      R, S, T, U;
    WORD64      M[3];

    if (final)
    {
        A = state->hs_State.small[0] = rotr64(state->hs_State.small[0], 1);
        B = state->hs_State.small[1] = rotr64(state->hs_State.small[1], 1);
        C = state->hs_State.small[2] = rotr64(state->hs_State.small[2], 1);
        D = state->hs_State.small[3] = rotr64(state->hs_State.small[3], 1);
        E = state->hs_State.small[4] = rotr64(state->hs_State.small[4], 1);
        F = state->hs_State.small[5] = rotr64(state->hs_State.small[5], 1);
        G = state->hs_State.small[6] = rotr64(state->hs_State.small[6], 1);
        P = state->hs_State.small[7] = rotr64(state->hs_State.small[7], 1);
        Q = state->hs_State.small[8] = rotr64(state->hs_State.small[8], 1);
    }
    else
    {
        state->hs_State.small[0] = rotr64(A = state->hs_State.small[0], 1);
        state->hs_State.small[1] = rotr64(B = state->hs_State.small[1], 1);
        state->hs_State.small[2] = rotr64(C = state->hs_State.small[2], 1);
        state->hs_State.small[3] = rotr64(D = state->hs_State.small[3], 1);
        state->hs_State.small[4] = rotr64(E = state->hs_State.small[4], 1);
        state->hs_State.small[5] = rotr64(F = state->hs_State.small[5], 1);
        state->hs_State.small[6] = rotr64(G = state->hs_State.small[6], 1);
        state->hs_State.small[7] = rotr64(P = state->hs_State.small[7], 1);
        state->hs_State.small[8] = rotr64(Q = state->hs_State.small[8], 1);
    }

    _512_MSG_EXP(W, state->hs_DataBuffer);

    _512_STEP_9(A, B, C, D, E, F, G, P, Q, 0);
    _512_STEP_9(A, B, C, D, E, F, G, P, Q, 9);
    _512_STEP_9(A, B, C, D, E, F, G, P, Q,18);
    _512_STEP_9(A, B, C, D, E, F, G, P, Q,27);
    _512_STEP_4(A, B, C, D, E, F, G, P, Q,36);

    state->hs_State.large[0] ^= F;
    state->hs_State.large[1] ^= G;
    state->hs_State.large[2] ^= P;
    state->hs_State.large[3] ^= Q;
    state->hs_State.large[4] ^= A;
    state->hs_State.large[5] ^= B;
    state->hs_State.large[6] ^= C;
    state->hs_State.large[7] ^= D;
    state->hs_State.large[8] ^= E;
}
